#!/usr/bin/env python3
# Demo that makes one Crazyflie take off 30cm above the first controller found
# Using the controller trigger it is then possible to 'grab' the Crazyflie
# and to make it move.
import csv
import sys
import time

import math
import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.crazyflie.syncLogger import SyncLogger
import matplotlib.pyplot as plt
from pathlib import Path
import numpy as np
from scipy.io import savemat
import matlab.engine

# URI to the Crazyflie to connect to
uri = 'radio://0/90/2M'

pos_list = []
rot_list = []

start_time = time.time_ns()


def wait_for_position_estimator(scf):
    print('Waiting for estimator to find position...')

    log_config = LogConfig(name='Kalman Variance', period_in_ms=500)
    log_config.add_variable('kalman.varPX', 'float')
    log_config.add_variable('kalman.varPY', 'float')
    log_config.add_variable('kalman.varPZ', 'float')

    var_y_history = [1000] * 10
    var_x_history = [1000] * 10
    var_z_history = [1000] * 10

    threshold = 0.001

    with SyncLogger(scf, log_config) as logger:
        for log_entry in logger:
            data = log_entry[1]

            var_x_history.append(data['kalman.varPX'])
            var_x_history.pop(0)
            var_y_history.append(data['kalman.varPY'])
            var_y_history.pop(0)
            var_z_history.append(data['kalman.varPZ'])
            var_z_history.pop(0)

            min_x = min(var_x_history)
            max_x = max(var_x_history)
            min_y = min(var_y_history)
            max_y = max(var_y_history)
            min_z = min(var_z_history)
            max_z = max(var_z_history)

            # print("{} {} {}".
            #       format(max_x - min_x, max_y - min_y, max_z - min_z))

            if (max_x - min_x) < threshold and (
                    max_y - min_y) < threshold and (
                    max_z - min_z) < threshold:
                break



def reset_estimator(scf):
    cf = scf.cf
    cf.param.set_value('kalman.resetEstimation', '1')
    time.sleep(0.1)
    cf.param.set_value('kalman.resetEstimation', '0')

    wait_for_position_estimator(cf)


def position_callback(timestamp, data, logconf):
    x = data['kalman.stateX']
    y = data['kalman.stateY']
    z = data['kalman.stateZ']

    time_since_start = (time.time_ns() - start_time) / (1000*1000*1000)
    # print('{} pos: ({}, {}, {})'.format(time_since_start, x, y, z))

    pos_list.append([time_since_start, x, y, z])


def rotation_callback(timestamp, data, logconf):
    q0 = data['kalman.q0']
    q1 = data['kalman.q1']
    q2 = data['kalman.q2']
    q3 = data['kalman.q3']

    time_since_start = (time.time_ns() - start_time) / (1000 * 1000 * 1000)

    euler_angles = quaternion_to_euler(q0, q1, q2, q3)
    yaw = euler_angles[0]
    pitch = euler_angles[1]
    roll = euler_angles[2]

    print('{} rot: ({}, {}, {})'.format(time_since_start, yaw, pitch, roll))

    rot_list.append([time_since_start, yaw, pitch, roll])


def quaternion_to_euler(q0, q1, q2, q3):
    """
    Convert a quaternion to Euler angles (yaw, pitch, roll) in radians.

    Args:
        q0 (float): The scalar (real) part of the quaternion.
        q1 (float): The x (imaginary) part of the quaternion.
        q2 (float): The y (imaginary) part of the quaternion.
        q3 (float): The z (imaginary) part of the quaternion.

    Returns:
        tuple: A tuple containing the yaw, pitch, and roll angles in radians.

    Generated by ChatGPT :)
    """
    # roll (x-axis rotation)
    sinr_cosp = 2 * (q0 * q1 + q2 * q3)
    cosr_cosp = 1 - 2 * (q1**2 + q2**2)
    roll = math.atan2(sinr_cosp, cosr_cosp)

    # pitch (y-axis rotation)
    sinp = 2 * (q0 * q2 - q3 * q1)
    if abs(sinp) >= 1:
        pitch = math.copysign(math.pi / 2, sinp)  # use 90 degrees if out of
        # range
    else:
        pitch = math.asin(sinp)

    # yaw (z-axis rotation)
    siny_cosp = 2 * (q0 * q3 + q1 * q2)
    cosy_cosp = 1 - 2 * (q2**2 + q3**2)
    yaw = math.atan2(siny_cosp, cosy_cosp)

    return (yaw, pitch, roll)


def merge_time_series(rotations_list, positions_list):
    """
    Merge two asynchronous time series based on closest time values.

    The function takes two lists of data, `rotations_list` and `positions_list`,
    where each list contains a sequence of data points with the first element
    representing a timestamp and the remaining elements representing data
    values. The function merges the two lists based on the closest timestamp
    values and returns a list of merged data points, with each point
    represented as a list containing timestamp followed by positional and
    rotational data.

    Args:
        rotations_list (list): A list of rotations data points, where each point
            contains a timestamp and yaw, pitch, and roll values.
        positions_list (list): A list of positions data points, where each point
            contains a timestamp and x, y, and z values.

    Returns:
        merged_list (list): A list of merged data points, where each point
            contains a timestamp and corresponding positional and rotational
            data. If there are no overlapping timestamps between the two
            input lists, the output list will be empty.

    Generated by ChatGPT :)
    """
    # Initialize empty list to store merged data points
    merged_list = []

    # Loop through each position data point
    for pos_point in positions_list:
        # Get the timestamp for the current position data point
        pos_time = pos_point[0]

        # Initialize variable to store closest rotation data point
        closest_rot_point = None
        closest_rot_time_diff = float('inf')

        # Loop through each rotation data point
        for rot_point in rotations_list:
            # Get the timestamp for the current rotation data point
            rot_time = rot_point[0]

            # Calculate the absolute difference between the position timestamp
            # and the rotation timestamp
            time_diff = abs(rot_time - pos_time)

            # Check if the current rotation point is closer in time than the
            # previously closest one
            if time_diff < closest_rot_time_diff:
                closest_rot_point = rot_point
                closest_rot_time_diff = time_diff

        # Check if a closest rotation point was found
        if closest_rot_point is not None:
            # Combine the position and rotation data points into a single
            # merged point
            merged_point = [pos_time] + pos_point[1:] + closest_rot_point[1:]

            # Add the merged point to the list of merged data points
            merged_list.append(merged_point)

    # Return the list of merged data points
    return merged_list


def extract_along_dimension(list_in, dimension_index):
    list_out = []

    for i in range(0, len(list_in)):
        list_out.append(list_in[i][dimension_index])

    return list_out


def start_position_printing(scf):
    log_conf = LogConfig(name='Rotation', period_in_ms=10)
    log_conf.add_variable('kalman.q0', 'float')
    log_conf.add_variable('kalman.q1', 'float')
    log_conf.add_variable('kalman.q2', 'float')
    log_conf.add_variable('kalman.q3', 'float')

    log_conf2 = LogConfig(name='Position', period_in_ms=10)
    log_conf2.add_variable('kalman.stateX', 'float')
    log_conf2.add_variable('kalman.stateY', 'float')
    log_conf2.add_variable('kalman.stateZ', 'float')

    scf.cf.log.add_config(log_conf)
    scf.cf.log.add_config(log_conf2)

    log_conf.data_received_cb.add_callback(rotation_callback)
    log_conf.start()

    log_conf2.data_received_cb.add_callback(position_callback)
    log_conf2.start()


def vector_substract(v0, v1):
    return [v0[0] - v1[0], v0[1] - v1[1], v0[2] - v1[2]]


def vector_add(v0, v1):
    return [v0[0] + v1[0], v0[1] + v1[1], v0[2] + v1[2]]


def fly_in_rectangle(cf, forward_dist, sideways_dist, move_time, wait_time):
    # left
    cf.high_level_commander.go_to(
        0, sideways_dist, 0, 0, move_time, relative=True
    )
    time.sleep(wait_time)

    # forward
    cf.high_level_commander.go_to(
        forward_dist, 0, 0, 0, move_time, relative=True
    )
    time.sleep(wait_time)

    # right
    cf.high_level_commander.go_to(
        0, -sideways_dist, 0, 0, move_time, relative=True
    )
    time.sleep(wait_time)

    # backwards
    cf.high_level_commander.go_to(
        -forward_dist, 0, 0, 0, move_time, relative=True
    )
    time.sleep(wait_time)


def run_sequence(scf):
    cf = scf.cf

    setpoint = [0, 0, 2]

    print("Initiating sequence")

    setpoint = [0, 0, .5]

    z_target = 0.66
    move_dist = 0.2
    move_vel = 0.2

    cf.param.set_value('commander.enHighLevel', '1')
    cf.high_level_commander.takeoff(z_target, 1)
    time.sleep(3)

    for i in range(0, 1):
        fly_in_rectangle(cf, 0.5, 0.6, 2.5, 4)

    time.sleep(3)
    # land
    cf.high_level_commander.land(0, z_target / move_vel)
    time.sleep(3)
    print("done")


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:

        # reset_estimator(scf) # if everything starts breaking, comment out
        # this line and make sure the position looks good in cfclient. -Austin
        start_position_printing(scf)
        run_sequence(scf)

        print(rot_list)
        print(pos_list)

        merged_list = merge_time_series(rot_list, pos_list)

        print(merged_list)

        times = extract_along_dimension(merged_list, 0)
        xs = extract_along_dimension(merged_list, 1)
        ys = extract_along_dimension(merged_list, 2)
        zs = extract_along_dimension(merged_list, 3)
        yaws = extract_along_dimension(merged_list, 4)
        pitches = extract_along_dimension(merged_list, 5)
        rolls = extract_along_dimension(merged_list, 6)

        """
        fig, axs = plt.subplots(1, 2)

        axs[0].plot(times, xs, label='X')
        axs[0].plot(times, ys, label='Y')
        axs[0].plot(times, zs, label='Z')

        # Add title and labels for the left subplot
        axs[0].set_title('Left Subplot')
        axs[0].set_xlabel('X-axis')
        axs[0].set_ylabel('Y-axis')

        # Add a legend to the left subplot
        axs[0].legend()

        axs[1].plot(times, yaws, label='Yaw')
        axs[1].plot(times, pitches, label='Pitch')
        axs[1].plot(times, rolls, label='Roll')

        # Add title and labels for the right subplot
        axs[1].set_title('Right Subplot')
        axs[1].set_xlabel('X-axis')
        axs[1].set_ylabel('Y-axis')

        # Add a legend to the right subplot
        axs[1].legend()

        # Adjust the layout of the subplots
        plt.tight_layout()

        plt.show()
        """

        matlab_filepath = Path("./../controls/Sim-nonlinear-quad-example/"
                               "Sim-nonlinear-quad-example")
        csv_filename = "measured_flight_path.mat"

        full_path = matlab_filepath.joinpath(csv_filename)

        merged_list_np = np.array(merged_list).transpose()

        mdic = {"measured_path": merged_list_np, "label": 'experiment'}
        savemat(full_path, mdic)
